"""
Helpers
=======

Useful functions and classes that don't really fit elsewhere.
"""

def is_update_modifier(doc):
    """Return true if and only if `doc` is an update modifier.

    This function only checks the first key it finds.  `doc` is assumed to
    be either a valid update modifier or a replacement document.  The result of
    passing a bad modifier or document is undefined.
    """
    if not doc:
        return False

    # To save time just check the first key.
    key = doc.iterkeys().next()
    return key.startswith('$')


def value_or_result(potential):
    """Return `potential` or its return value.

    If `potential` is callable then the result of calling it without arguments
    is returned instead.

    """
    if callable(potential):
        return potential()
    return potential


class ClassDefault(object):
    """A sentinel value signaling that the class default should be used.

    This is used as default value for keyword arguments so that a default can
    be set during class declaration, but the value can still be explicitly
    specified when the function is called.
    """


class ConversionDict(dict):
    """A :class:`dict` subclass that wraps contents on lookup.

    >>> cd = ConversionDict({'x': 5, 'y': 6}, {'x': str})
    >>> cd['x']
    '5'
    >>> cd.x
    '5'
    >>> cd['y']
    6
    >>> cd.y
    6

    :param content: is the dictionary content.
    :param conversions: is a dictionary containing conversions to be applied on
        lookup.  If no conversion is present in `conversions` the value is
        returned unchanged.
    """

    NONKEY_ATTRS = set(['_conversions'])
    """Attributes which should not be treated as keys in the dictionary.

    Classes extending :class:`ConversionDict` should add the names of any
    attributes they wish to keep that should not be set into the underlying
    :class:`dict`.
    """

    def __init__(self, content, conversions):
        dict.__init__(self, content)
        self._conversions = conversions

    def __getitem__(self, key):
        value = dict.__getitem__(self, key)
        conversion = self.__get_conversion(key)
        return self.__convert_value(value, conversion)

    @staticmethod
    def __convert_value(value, conversion):
        """Convert `value` using `conversion`.

        If `conversion` is ``None`` then `value` is returned unchanged.
        """
        if conversion is None:
            return value

        if isinstance(conversion, dict):
            return ConversionDict(value, conversion)

        if isinstance(value, list):
            return [conversion(x) for x in value]

        return conversion(value)

    def __get_conversion(self, key):
        if self._conversions:
            return self._conversions.get(key)

    def __getattr__(self, name):
        """Get the `name` attribute.

        If `name` is a :class:`dict` then it will be wrapped in a new
        :class:`ConversionDict`.  This allows chained dot notation to reference
        items in embedded dictionaries.
        """
        if name in self:
            return self.__getitem__(name)

        raise AttributeError('{0.__name__!r} has no attribute {1!r}.'.format(
            type(self), name))

    def __setattr__(self, name, value):
        """Set an attribute.

        Put an item in this dictionary unless `name` is a contained in
        :data:`NONKEY_ATTRS`.  Or some other attribute already set on either
        the instance or the class (which allows the tests to dingus out methods
        on instances).
        """
        if name in self.NONKEY_ATTRS or name in dir(self):
            object.__setattr__(self, name, value)
        else:
            self.__setitem__(name, value)

    def iteritems(self):
        """Iterate the items in this :class:`ConversionDict`.

        The underlying :meth:`dict.iteritems` must be wrapped such that the
        values returned by this generator are converted to appropriate types.
        """
        for key, value in dict.iteritems(self):
            conversion = self.__get_conversion(key)
            yield key, self.__convert_value(value, conversion)


    def itervalues(self):
        """Iterate the values in this :class:`ConversionDict`.

        The values are converted to the appropriate type before returning.
        """
        for key, value in self.iteritems():
            yield value

    def items(self):
        """Return a list of the items in this :class:`ConversionDict`.
        """
        return [item for item in self.iteritems()]

    def values(self):
        """Return a list of the values in this :class:`ConversionDict`.
        """
        return [value for value in self.itervalues()]



"""
Connection
==========
"""
import warnings

import pymongo

from scalymongo.document import get_concrete_classes


class Connection(pymongo.Connection):
    """A connection to a MongoDB database.

    This is a wrapper for a :class:`pymongo.connection.Connection`.
    """

    def connect_document(self, document):
        """Connect a document by creating a new type and injecting the
        connection.
        """
        attrs = {
            'connection': self,
            'database': self[document.__database__],
            'collection': self[document.__database__][document.__collection__],
        }
        return type('Connected{0}'.format(document.__name__),
                    (document,),
                    attrs)

    @property
    def models(self):
        return DocumentProxy(self, get_concrete_classes())


class DocumentProxy(object):
    """A proxy object for accessing or creating :class:`Document` models."""

    def __init__(self, connection, registered):
        self.connection = connection
        self.registered = {}
        for cls in registered:
            if cls.__name__ in self.registered:
                warnings.warn(
                    'Multiple models have been found with the name {0}.'
                    ' The result of connection.models[{0}] will be undefined.'
                    .format(repr(cls.__name__)))
            self.registered[cls.__name__] = cls

    def __getitem__(self, name):
        cls = self._find_document_class(name)
        if not cls:
            raise KeyError('Unknown document {0}'.format(repr(name)))
        return cls

    def __getattr__(self, name):
        cls = self._find_document_class(name)
        if not cls:
            raise AttributeError('Unknown document {0}'.format(repr(name)))
        return cls

    def _find_document_class(self, name):
        cls = self.registered.get(name)
        if not cls:
            return None
        return self.connection.connect_document(cls)

    def __iter__(self):
        """A generator for all models registered on this connection.
        """
        for value in self.registered.itervalues():
            yield self.connection.connect_document(value)



from optparse import OptionParser

from scalymongo.connection import Connection


def parse_arguments():
    """Parse and validate the arguments.

    This returns a tuple like ``(options, module_name, endpoint)``.

    """
    parser = OptionParser()
    parser.usage = '%prog [options] MODULE ENDPOINT'
    parser.add_option(
        '--background', action='store_true',
        help='create indexes as a non-blocking operation [default]',
    )
    parser.add_option(
        '--no-background', action='store_false', dest='background',
        help='disable background index creation',
    )
    parser.set_defaults(background=True)

    options, arguments = parser.parse_args()

    if len(arguments) != 2:
        parser.print_help()
        exit(1)

    module_name, endpoint = arguments
    return options, module_name, endpoint


def main():
    options, module_name, endpoint = parse_arguments()

    # Import the models so that they get registered with Connection.
    __import__(module_name)

    connection = Connection(endpoint)
    ensure_indexes(connection, options)


def ensure_indexes(connection, options):
    for model in connection.models:
        model.ensure_indexes(background=options.background)


if __name__ == '__main__': # pragma: no cover
    main()



"""
Structure Walker
================

A utility used to aid in structure validation.
"""
from inspect import isclass

from scalymongo.errors import ValidationError


class StructureWalker(object):
    """A helper class to recurse a :class:`dict`-like object in accordance with
    a structure.

    :param field_translator: should be function mapping the ``value`` and
        ``type_`` to the new value for a key.

    """

    def __init__(self, field_validator):
        self.field_validator = field_validator

    def walk_dict(self, body, structure, path=None):
        """Validate a dictionary in accordance with `structure`.

        A :class:`ValidationError` is raised if any fields in `body` are
        not present in `structure`.
        """
        _check_for_unknown_fields(body, structure, path)

        for field, sub_structure in structure.iteritems():
            if isclass(field):
                field_type = field
                # For structures like {<TYPE>: {<STRUCT>}} iterate values
                # in the body with keys of <TYPE> and verify each against
                # <STRUCT>.
                for key, value in body.iteritems():
                    if isinstance(key, field_type):
                        self._recurse_or_validate_field(
                            value, sub_structure, _join(path, key))

            if field in body:
                self._recurse_or_validate_field(
                    body[field], sub_structure, _join(path, field))

    def _recurse_or_validate_field(self, value, sub_structure, path):
        if isinstance(sub_structure, list):
            assert len(sub_structure) == 1
            # Validate each value in the list against the specified content
            # type.
            for i, value in enumerate(value):
                self._recurse_or_validate_field(
                    value, sub_structure[0], _join(path, i))
            return

        if isinstance(sub_structure, dict):
            self.walk_dict(value, sub_structure, path)
            return

        self.field_validator(path, value, sub_structure)


def _check_for_unknown_fields(body, structure, path):
    """Check `body` for any keys not present in `structure`.

    This only checks the first level of keys.  Any keys from :class:`dict`s in
    the `body`\ 's values will not be checked.
    """
    type_keys = tuple([key for key in structure if isclass(key)])
    existing_fields = set([key for key in body if not isclass(key)])
    unknown_fields = existing_fields.difference(structure.keys())
    # If there are valid types for a key filter out unknown fields that match a
    # type.
    if type_keys:
        unknown_fields = [key for key in unknown_fields
                          if not isinstance(key, type_keys)]

    if unknown_fields:
        unknown_fields = ', '.join([repr(field) for field in unknown_fields])
        if path:
            err = ('Encountered field(s), in subdocument at {0},'
                   ' not present in structure: {1}'.format(
                       path, unknown_fields))
        else:
            err = 'Encountered field(s) not present in structure: {0}'.format(
                unknown_fields)

        raise ValidationError(err)


def _join(head, tail):
    """Join `head` and `tail` with a dot.

    If head is ``None`` only `tail` is returned.
    """
    if head is None:
        return tail
    return '{0}.{1}'.format(head, tail)



# -*- coding: utf-8 -*-


def _make_cursor_wrapper_method(method_name):
    """Make a wrapper method for `method_name`.

    `method_name` should be the name of a method on
    :class:`pymongo.cursor.Cursor` which returns another
    :class:`pymongo.cursor.Cursor`.

    The wrapper method will pass along any positional and/or keyword arguments
    to the underlying method of the wrapped cursor and wrap the result in a
    :class:`Cursor` so that it will still yield ScalyMongo documents.

    """
    def _wrapped_method(self, *args, **kwargs):
        method = getattr(self._Cursor__wrapped_cursor, method_name)
        returned = method(*args, **kwargs)
        return Cursor(returned, self._Cursor__document_type)

    return _wrapped_method


class Cursor(object):
    """Wrapper for :class:`pymongo.cursor.Cursor` objects.

    This class is a thin wrapper which takes results returned by the underlying
    cursor and wraps them in the appropriate
    :class:`~scalymongo.document.Document` subclass.
    """

    def __init__(self, wrapped_cursor, document_type):
        self.__wrapped_cursor = wrapped_cursor
        self.__document_type = document_type

    def __getitem__(self, index):
        """Get the item(s) at `index`.

        :param index: may also be a slice.
        """
        returned = self.__wrapped_cursor[index]
        # If the index is a slice then the result is a new cursor with the skip
        # and limit already applied.
        if isinstance(index, slice):
            return Cursor(returned, self.__document_type)
        return self.__document_type(returned)

    def next(self):
        """Return the next document for this cursor.
        """
        return self.__document_type(self.__wrapped_cursor.next())

    def __iter__(self):
        return self

    # Wrap all methods that return a pymongo.cursor.Cursor.
    batch_size = _make_cursor_wrapper_method('batch_size')
    clone = _make_cursor_wrapper_method('clone')
    hint = _make_cursor_wrapper_method('hint')
    limit = _make_cursor_wrapper_method('limit')
    max_scan = _make_cursor_wrapper_method('max_scan')
    rewind = _make_cursor_wrapper_method('rewind')
    skip = _make_cursor_wrapper_method('skip')
    sort = _make_cursor_wrapper_method('sort')
    where = _make_cursor_wrapper_method('where')

    def __getattr__(self, attr):
        """All other methods and properties are those of the wrapped cursor.
        """
        return getattr(self.__wrapped_cursor, attr)



import pprint

from scalymongo import Document, Connection

class FindExample(Document):
    structure = {
        'name': basestring,
        'age': int,
    }

    __database__ = 'test'
    __collection__ = __file__


def find():
    return conn.models.FindExample.find()

def render(cursor):
    print '=' * 80
    f = list(cursor)
    print [type(x) for x in f]
    pprint.pprint(f)

def main():

    docs = [
        {'name': 'Alice', 'age': 32},
        {'name': 'Bob', 'age': 32},
        {'name': 'Carl', 'age': 41},
        {'name': 'Donna', 'age': 35},
    ]
    docs = [conn.models.FindExample(doc) for doc in docs]
    for doc in docs:
        doc.save()

    print 'all'
    render(find())
    print '[2:3]'
    render(find()[2:3])
    print 'limit(1)'
    render(find().limit(1))
    print '[0:1]'
    render(find()[0:1])
    print 'skip(2)'
    render(find().skip(2))


if __name__ == '__main__':
    conn = Connection('localhost')
    conn.models.FindExample.remove({})
    main()
    conn.models.FindExample.remove({})



class UnsafeBehaviorError(Exception):
    pass


class GlobalQueryException(Exception):
    pass


class SchemaError(ValueError):
    pass


class ValidationError(Exception):
    pass


class ModifyFailedError(Exception):
    pass
